/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: Greedy.cpp                                                                 *
 * @author: Xu Le, He Jianfan                                                        *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#include "Greedy.h"

#define REMOVE_ACCESS_NODE     for (itS = servers.begin(); itS != servers.end(); ++itS) { if (*itS == cspRank[curIdx]) { servers.erase(itS); break; } }
#define REMOVE_CHANGED_NODE    for (itS = servers.begin(); itS != servers.end(); ++itS) { if (*itS == lastChangedNode) { servers.erase(itS); break; } }

extern int log_level;

extern uint32_t networkNodeNum;
extern uint32_t linkNum;
extern uint32_t consumptionNodeNum;
extern uint32_t hardwareLevelNum;

extern std::vector<int> priceTable;
extern std::vector<int> outcapTable;
extern std::vector<int> deployTable;
extern std::vector<int> accessTable;
extern std::vector<int> demandTable;
extern std::unordered_map<int, int> provideTable;

extern std::list<int> curBestServers;

Greedy::Greedy(TopologyScanner& _scanner) : scanner(_scanner), progress(Progress::ZOM),
	minusPlusTimes(0), MAX_MINUS_PLUS_TIMES(networkNodeNum/600+1), MAX_TRY_TIMES_IN_EX(15), curIdx(0), curNBIdx(0),
	hasExecutedZOP(false), exchangeSwitch(false), inDecreaseProgress(false), inIncreaseProgress(false), inExchangeProgress(false), inRemovingProgress(false),
	kDecreasableNum(networkNodeNum/20), kIncreasableNum(networkNodeNum/20), kExchangeableNum(networkNodeNum/20), kExchangeableCN(networkNodeNum/20), kRemovableNum(5)
{
	deployed = new bool[networkNodeNum];
	cspCN = static_cast<uint32_t>(scanner.possibleServers.size());
	cspRank = new int[cspCN];
	tryDecreaseTimes = static_cast<int>(networkNodeNum) / 6; // needn't to be changed
	tryIncreaseTimes = static_cast<int>(networkNodeNum) / 6; // needn't to be changed

	decreasableArray.resize(kDecreasableNum);
	increasableArray.resize(kIncreasableNum);
	exchangeableArray.resize(kExchangeableNum);
	neighborArray.resize(1024);
	exchangeableCArray.resize(kExchangeableCN);
	removableArray.resize(kRemovableNum);
}

Greedy::~Greedy()
{
	delete []deployed;
	delete []cspRank;
}

/** API function of CDN server list initialization required by deploy.cpp. */
void Greedy::initialize(std::list<int>& servers)
{
	printf("====================    Enter greedy category    ====================\n");

	uint32_t i = 0;

	progress = Progress::ZOM;
	minusPlusTimes = 0;
	curIdx = 0;
	curNBIdx = 0;
	hasExecutedZOP = false;
	exchangeSwitch = false;
	inDecreaseProgress = false;
	inIncreaseProgress = false;
	inExchangeProgress = false;
	inRemovingProgress = false;
	tryDecreaseTimes = static_cast<int>(networkNodeNum) / 6; // needn't to be changed
	tryIncreaseTimes = static_cast<int>(networkNodeNum) / 6; // needn't to be changed
	for (i = 0; i < networkNodeNum; ++i)
		deployed[i] = false;
	deltaPrice.resize(hardwareLevelNum);
	deltaPrice[0] = priceTable[0];
	for (i = 1; i < hardwareLevelNum; ++i)
		deltaPrice[i] = priceTable[i] - priceTable[i-1];
	exchangeTimes.clear();

	for (i = 0; i < cspCN; ++i)
		cspRank[i] = scanner.possibleServers[cspCN-i-1];

	info_log("====================    Progress::ZOM(1) starts    ====================\n");

	for (i = 1; i < cspCN; ++i) // the least rank node has been removed
	{
		servers.push_back(cspRank[i]);
		deployed[cspRank[i]] = true;
	}

	for (std::set<int>::iterator itDef = scanner.definiteServers.begin(); itDef != scanner.definiteServers.end(); ++itDef)
	{
		servers.push_back(*itDef);
		deployed[*itDef] = true;
	}

	lastChangedNode = cspRank[curIdx];
	// set initial total fee to infinity so that optimize() progress can execute normally, though the lease rank node will be removed without examination
	lastResultFee = INF;
}

/**
 * @brief API function of CDN server list optimization required by deploy.cpp.
 * 
 * 1. initial state is Progress::ZOM and bool variable hasExecutedZOP == false, thus starts at line 118, when it is time to exit this state, then go to line 127,137;
 * 2. when state switched to Progress::ZOP, starts at line 163, when it is time to exit this state, set bool variable hasExecutedZOP = true, then go to line 180,184;
 * 3. when state switched back to Progress::ZOM, starts at line 86, when it is time to exit this state, then go to line 111;
 */
bool Greedy::optimize(std::list<int>& servers, std::vector<int>& serversOutputFlux, long justRentalFee, long purchaseFee, long deploymentFee)
{
	uint32_t i = 0;
	long justResultFee = justRentalFee + purchaseFee + deploymentFee;

	switch (progress)
	{
	case Progress::ZOM:
	{
		if (hasExecutedZOP) // usually will enter this part
		{
			if (lastResultFee <= justResultFee || justRentalFee == -1) // restore the most recently removed node
			{
				servers.push_front(lastChangedNode);
				deployed[lastChangedNode] = true;
			}
			else
				lastResultFee = justResultFee;

			while (!deployed[cspRank[curIdx]] && curIdx < static_cast<int>(cspCN-1))
				++curIdx;
			if (curIdx < static_cast<int>(cspCN-1)) // quit above while loop due to the first condition
			{
				REMOVE_ACCESS_NODE
				deployed[cspRank[curIdx]] = false;
				lastChangedNode = cspRank[curIdx++];
			}
			else
			{
				if (++minusPlusTimes < MAX_MINUS_PLUS_TIMES)
					curIdx = -1; // prepare to start ZOP progress
				else
					curIdx = -2; // prepare to start DL progress
				hasExecutedZOP = false;
			}
		}
		else if (curIdx > -1) // only the initial state ZOM will enter this part
		{
			if (lastResultFee <= justResultFee || justRentalFee == -1) // restore the most recently removed node
			{
				servers.push_front(lastChangedNode);
				deployed[lastChangedNode] = true;
			}
			else
				lastResultFee = justResultFee;

			if (++curIdx >= static_cast<int>(cspCN-1))
				curIdx = -1; // prepare to start ZOP progress
			else
			{
				REMOVE_ACCESS_NODE
				deployed[cspRank[curIdx]] = false;
				lastChangedNode = cspRank[curIdx];
			}
		}
		else if (curIdx > -2) // actually here is a transitional state when the next loop will switch to Progress::ZOP
		{
			lastResultFee = justResultFee;

			info_log("Total fee after Progress::ZOM(%u) is %ld\n\n", minusPlusTimes, lastResultFee);
			info_log("====================    Progress::ZOP(%u) starts    ====================\n", minusPlusTimes);

			curIdx = 0; // return to 0, loop restarts again
			while (deployed[cspRank[curIdx]])
				++curIdx;

			servers.push_front(cspRank[curIdx]);
			deployed[cspRank[curIdx]] = true;
			lastChangedNode = cspRank[curIdx++];

			progress = Progress::ZOP;
		}
		else // actually here is a transitional state when the next loop will switch to Progress::DL
		{
			lastResultFee = justResultFee;

			servers = curBestServers;
			info_log("Total fee after Progress::ZOM(%u) is %ld\n\n", minusPlusTimes, lastResultFee);
			info_log("====================    Progress::DL starts    ====================\n");

			curIdx = 0;
			progress = Progress::DL;
		}
		break;
	}
	case Progress::ZOP:
	{
		if (!hasExecutedZOP) // each time when state has just switched to zero order plus, hasExecutedZOP is always false, see line 116
		{
			if (lastResultFee <= justResultFee || justRentalFee == -1) // remove the most recently added node
			{
				servers.pop_front();
				deployed[lastChangedNode] = false;
			}
			else
				lastResultFee = justResultFee;

			while (deployed[cspRank[curIdx]] && curIdx < static_cast<int>(cspCN-1))
				++curIdx;
			if (curIdx < static_cast<int>(cspCN-1)) // quit above while loop due to the first condition
			{
				servers.push_front(cspRank[curIdx]);
				deployed[cspRank[curIdx]] = true;
				lastChangedNode = cspRank[curIdx++];
			}
			else
				hasExecutedZOP = true;
		}
		else
		{
			lastResultFee = justResultFee;

			info_log("Total fee after Progress::ZOP(%u) is %ld\n\n", minusPlusTimes, lastResultFee);
			info_log("====================    Progress::ZOM(%u) starts    ====================\n", minusPlusTimes+1);

			curIdx = 0; // return to 0, loop restarts again
			while (!deployed[cspRank[curIdx]])
				++curIdx;

			REMOVE_ACCESS_NODE
			deployed[cspRank[curIdx]] = false;
			lastChangedNode = cspRank[curIdx++];

			progress = Progress::ZOM;
		}
		break;
	}
	case Progress::DL:
	{
		if (!inDecreaseProgress)
		{
			inDecreaseProgress = true;
			// info_log("====================    Progress::DL starts    ====================\n");
			// info_log("DL --- justResultFee: %ld\n", justResultFee);
			--tryDecreaseTimes;

			std::vector<struct OverFluxTuple> sortOverFluxFee;
			sortOverFluxFee.reserve(servers.size());
			for (itS = servers.begin(), i = 0; itS != servers.end(); ++itS, ++i)
			{
				int curLevel = fitLevel(serversOutputFlux[i]);
				if (curLevel > 0 && serversOutputFlux[i] < outcapTable[curLevel])
				{
					int overFlux = serversOutputFlux[i] - outcapTable[curLevel-1];
					sortOverFluxFee.emplace_back(deltaPrice[curLevel]/overFlux, *itS, curLevel);
				}
			}
			std::sort(sortOverFluxFee.begin(), sortOverFluxFee.end(), [](const struct OverFluxTuple& lhs, const struct OverFluxTuple& rhs) { return lhs.unitFee > rhs.unitFee; });
			decreasableNum = std::min(kDecreasableNum, static_cast<uint32_t>(sortOverFluxFee.size()));
			for (i = 0; i < decreasableNum; ++i)
			{
				decreasableArray[i].first = sortOverFluxFee[i].nodeID;
				decreasableArray[i].second = sortOverFluxFee[i].preLevel;
			}
			decreasableIdx = 0;
		}

		lastChangedNode = decreasableArray[decreasableIdx].first;
		int postLevel = decreasableArray[decreasableIdx++].second - 1;
		if (tryDecreaseTimes < 0)
		{
			printf("====================    Exit greedy category    ====================\n\n");
			return false;
		}

		scanner.graph.setOutputFluxUpbound(lastChangedNode, outcapTable[postLevel]);
		progress = Progress::DLSTEP;
		break;
	}
	case Progress::DLSTEP:
	{
		if (lastResultFee <= justResultFee || justRentalFee == -1) // revert to unrestricted output capacity
		{
			int preLevel = decreasableArray[decreasableIdx-1].second;
			scanner.graph.setOutputFluxUpbound(lastChangedNode, outcapTable[preLevel]);
			if (decreasableIdx < decreasableNum)
			{
				lastChangedNode = decreasableArray[decreasableIdx].first;
				int postLevel = decreasableArray[decreasableIdx].second - 1;
				scanner.graph.setOutputFluxUpbound(lastChangedNode, outcapTable[postLevel]);
				++decreasableIdx;
			}
			else
			{
				inDecreaseProgress = false;
				progress = networkNodeNum < 1000 ? Progress::IL : Progress::EX;
			}
		}
		else
		{
			lastResultFee = justResultFee;

			info_log("[%d]'s level decreased in Progress::DL\t---  %ld\n", lastChangedNode, justResultFee);

			inDecreaseProgress = false;
			progress = Progress::DL; // return back to Progress::DL and recalculate unit over flux fee
		}
		break;
	}
	case Progress::IL:
	{
		if (!inIncreaseProgress)
		{
			inIncreaseProgress = true;
			// info_log("====================    Progress::IL starts    ====================\n");
			// info_log("IL --- justResultFee: %ld\n", justResultFee);
			--tryIncreaseTimes;

			decreasableIdx = 0;

			std::vector<struct OverFluxTuple> sortOverFluxFee;
			sortOverFluxFee.reserve(servers.size());
			for (itS = servers.begin(), i = 0; itS != servers.end(); ++itS, ++i)
			{
				int curLevel = fitLevel(serversOutputFlux[i]);
				if (curLevel < static_cast<int>(hardwareLevelNum)-1 && serversOutputFlux[i] == outcapTable[curLevel])
				{
					int overFlux = serversOutputFlux[i] - outcapTable[curLevel-1];
					sortOverFluxFee.emplace_back(deltaPrice[curLevel]/overFlux, *itS, curLevel);
				}
			}
			std::sort(sortOverFluxFee.begin(), sortOverFluxFee.end(), [](const struct OverFluxTuple& lhs, const struct OverFluxTuple& rhs) { return lhs.unitFee < rhs.unitFee; });
			increasableNum = std::min(kIncreasableNum, static_cast<uint32_t>(sortOverFluxFee.size()));
			for (i = 0; i < increasableNum; ++i)
			{
				increasableArray[i].first = sortOverFluxFee[i].nodeID;
				increasableArray[i].second = sortOverFluxFee[i].preLevel;
			}
		}

		lastChangedNode = decreasableArray[decreasableIdx].first;
		int postDecreaseLevel = decreasableArray[decreasableIdx].second - 1;
		if (++decreasableIdx > decreasableNum)
		{
			inIncreaseProgress = false;
			progress = Progress::EX;
			return true;
		}

		increasableIdx = 0;
		if (lastChangedNode == increasableArray[increasableIdx].first)
			++increasableIdx;

		int curUpgradedNode = increasableArray[increasableIdx].first;
		int postIncreaseLevel = increasableArray[increasableIdx].second + 1;
		scanner.graph.setOutputFluxUpbound(lastChangedNode, outcapTable[postDecreaseLevel]);
		scanner.graph.setOutputFluxUpbound(curUpgradedNode, outcapTable[postIncreaseLevel]);
		++increasableIdx;
		progress = Progress::ILSTEP;
		break;
	}
	case Progress::ILSTEP:
	{
		int lastUpgradedNode = increasableArray[increasableIdx-1].first;
		if (lastResultFee <= justResultFee || justRentalFee == -1) // revert to unrestricted output capacity and unupgrade output capacity
		{
			int preIncreaseLevel = increasableArray[increasableIdx-1].second;
			scanner.graph.setOutputFluxUpbound(lastUpgradedNode, outcapTable[preIncreaseLevel]); // undo previous upgrade operation
			if (lastChangedNode == increasableArray[increasableIdx].first)
				++increasableIdx;
			if (increasableIdx < increasableNum) // try to upgrade the next increasable node
			{
				int curUpgradedNode = increasableArray[increasableIdx].first;
				int postIncreaseLevel = increasableArray[increasableIdx].second + 1;
				scanner.graph.setOutputFluxUpbound(curUpgradedNode, outcapTable[postIncreaseLevel]);
				++increasableIdx;
			}
			else
			{
				int preDecreaseLevel = decreasableArray[decreasableIdx-1].second;
				scanner.graph.setOutputFluxUpbound(lastChangedNode, outcapTable[preDecreaseLevel]);

				progress = Progress::IL; // return back to Progress::IL and try the next decreasable node
			}
		}
		else
		{
			lastResultFee = justResultFee;

			info_log("[%d]'s level decreased in Progress::IL\t---  %ld\n", lastChangedNode, justResultFee);
			info_log("[%d]'s level increased in Progress::IL\t---  %ld\n", lastUpgradedNode, justResultFee);

			inIncreaseProgress = false;
			progress = Progress::EX; // go to Progress::DL and try to decrease server's level in new scheme
		}
		break;
	}
	case Progress::EX:
	{
		if (!inExchangeProgress)
		{
			inExchangeProgress = true;
			// info_log("====================    Progress::EX starts    ====================\n");
			// info_log("EX --- justResultFee: %ld\n", justResultFee);

			std::vector<std::pair<double /* unit flux fee */, int /* node ID */> > sortUnitFluxFee;
			sortUnitFluxFee.reserve(servers.size());
			serverHashset.clear();
			for (itS = servers.begin(), i = 0; itS != servers.end(); ++itS, ++i)
			{
				serverHashset.insert(*itS);
				int curLevel = fitLevel(serversOutputFlux[i]);
				double totalNodeFee = priceTable[curLevel] + deployTable[*itS];
				sortUnitFluxFee.emplace_back(totalNodeFee/serversOutputFlux[i], *itS);
				// printf("%d[%.2f] ", *itS, totalNodeFee/serversOutputFlux[i]);
			}
			// printf("\n");
			std::sort(sortUnitFluxFee.begin(), sortUnitFluxFee.end(), std::greater<std::pair<double, int> >());
			exchangeableNum = std::min(kExchangeableNum, static_cast<uint32_t>(sortUnitFluxFee.size()));
			for (i = 0; i < exchangeableNum; ++i)
				exchangeableArray[i] = sortUnitFluxFee[i].second;
			exchangeableIdx = 0;
		}

		lastChangedNode = exchangeableArray[exchangeableIdx];
		if (++exchangeableIdx > exchangeableNum)
		{
			inExchangeProgress = false;
			progress = Progress::EX2;
			scanner.graph.setExchangeOperation(exchangeSwitch);
			return true;
		}

		neighborCN = 0;
		// printf("%d:  ", lastChangedNode);
		struct Arc *parc = NULL;
		for (parc = scanner.graph.NodeList[lastChangedNode].firstArc; parc != NULL; parc = parc->nextArc)
		{
			if (parc->dstID < scanner.mNetworkNodeNum
				&& deployTable[parc->dstID] <= deployTable[lastChangedNode]
				&& scanner.nodeLevel[parc->dstID] >= scanner.nodeLevel[lastChangedNode]
				&& serverHashset.find(parc->dstID) == serverHashset.end())
			{
				// printf("%d[%d] -> ", parc->dstID, scanner.nodeLevel[parc->dstID]);
				neighborArray[neighborCN++] = parc->dstID;
			}
		}
		// printf("\n");
		if (neighborCN > 0)
		{
			curNBIdx = 0;
			while (curNBIdx < neighborCN && ++exchangeTimes[uuid(lastChangedNode, neighborArray[curNBIdx])] > MAX_TRY_TIMES_IN_EX)
				++curNBIdx;

			if (curNBIdx < neighborCN)
			{
				REMOVE_CHANGED_NODE
				deployed[lastChangedNode] = false;
				servers.push_front(neighborArray[curNBIdx]);
				deployed[neighborArray[curNBIdx]] = true;
				++curNBIdx;

				progress = Progress::EXSTEP; // go to Progress::EXSTEP and try to exchange with neighbors one by one
			}
		}
		break;
	}
	case Progress::EXSTEP:
	{
		if (lastResultFee <= justResultFee || justRentalFee == -1) // try to exchange with another neighbor or restore the exchangeable node
		{
			servers.pop_front(); // remove the neighbor node added in previous exchange operation
			deployed[neighborArray[curNBIdx-1]] = false;
			while (curNBIdx < neighborCN && ++exchangeTimes[uuid(lastChangedNode, neighborArray[curNBIdx])] > MAX_TRY_TIMES_IN_EX)
				++curNBIdx;
			if (curNBIdx < neighborCN)
			{
				servers.push_front(neighborArray[curNBIdx]); // add another neighbor
				deployed[neighborArray[curNBIdx]] = true;
				++curNBIdx;
			}
			else
			{
				servers.push_front(lastChangedNode); // restore the exchangeable node removed in previous exchange operation
				deployed[lastChangedNode] = true;
				progress = Progress::EX; // return back to Progress::EX and try the next exchangeable node
			}
		}
		else
		{
			lastResultFee = justResultFee;

			info_log("node [%d] <-> [%d] in Progress::EX  \t---  %ld\n", lastChangedNode, neighborArray[curNBIdx-1], justResultFee);

			inExchangeProgress = false;
			progress = Progress::DL; // go to Progress::DL and try to decrease server's level in new scheme
		}
		break;
	}
	case Progress::EX2:
	{
		if (!inExchangeProgress)
		{
			inExchangeProgress = true;
			info_log("====================    Progress::EX2 starts    ====================\n");
			// info_log("EX2 --- justResultFee: %ld\n", justResultFee);

			exchangeableIdx = 0;

			if (exchangeSwitch)
			{
				scanner.graph.setExchangeOperation(false);
				std::vector<std::pair<int /* flux through */, int /* node ID */> > sortFluxThrough;
				sortFluxThrough.reserve(networkNodeNum-servers.size());
				for (int l = 0; l < scanner.mNetworkNodeNum; ++l)
					if (serverHashset.find(l) == serverHashset.end())
						sortFluxThrough.emplace_back(scanner.graph.fluxThroughNode[l], l);
				std::sort(sortFluxThrough.begin(), sortFluxThrough.end(), std::greater<std::pair<int, int> >());
				exchangeableCN = std::min(kExchangeableCN, static_cast<uint32_t>(sortFluxThrough.size()));
				for (i = 0; i < exchangeableCN; ++i)
					exchangeableCArray[i] = sortFluxThrough[i].second;
			}
			else
			{
				std::vector<std::pair<double /* unit flux fee */, int /* node ID */> > sortUnitFluxFee;
				sortUnitFluxFee.reserve(networkNodeNum-servers.size());
				for (int l = 0; l < scanner.mNetworkNodeNum; ++l)
				{
					if (serverHashset.find(l) == serverHashset.end())
					{
						double totalNodeFee = priceTable[scanner.nodeLevel[l]] + deployTable[l];
						sortUnitFluxFee.emplace_back(totalNodeFee/outcapTable[scanner.nodeLevel[l]], l);
					}
				}
				std::sort(sortUnitFluxFee.begin(), sortUnitFluxFee.end());
				exchangeableCN = std::min(kExchangeableCN, static_cast<uint32_t>(sortUnitFluxFee.size()));
				for (i = 0; i < exchangeableCN; ++i)
					exchangeableCArray[i] = sortUnitFluxFee[i].second;
			}
			exchangeSwitch = exchangeSwitch == false;
		}

		lastChangedNode = exchangeableArray[exchangeableIdx];
		if (++exchangeableIdx > exchangeableNum)
		{
			inExchangeProgress = false;
			progress = Progress::RM;
			return true;
		}

		exchangeableCIdx = 0;
		while (exchangeableCIdx < exchangeableCN && ++exchangeTimes[uuid(lastChangedNode, exchangeableCArray[exchangeableCIdx])] > MAX_TRY_TIMES_IN_EX)
			++exchangeableCIdx;

		if (exchangeableCIdx < exchangeableCN)
		{
			REMOVE_CHANGED_NODE
			deployed[lastChangedNode] = false;
			servers.push_front(exchangeableCArray[exchangeableCIdx]);
			deployed[exchangeableCArray[exchangeableCIdx]] = true;
			++exchangeableCIdx;

			progress = Progress::EX2STEP; // go to Progress::EX2STEP and try to exchange with exchangeable candidate nodes one by one
		}
		break;
	}
	case Progress::EX2STEP:
	{
		if (lastResultFee <= justResultFee || justRentalFee == -1) // try to exchange with exchangeable candidate node or restore the exchangeable node
		{
			servers.pop_front(); // remove the exchangeable candidate node added in previous exchange operation
			deployed[exchangeableCArray[exchangeableCIdx-1]] = false;
			while (exchangeableCIdx < exchangeableCN && ++exchangeTimes[uuid(lastChangedNode, exchangeableCArray[exchangeableCIdx])] > MAX_TRY_TIMES_IN_EX)
				++exchangeableCIdx;
			if (exchangeableCIdx < exchangeableCN)
			{
				servers.push_front(exchangeableCArray[exchangeableCIdx]); // add another exchangeable candidate node
				deployed[exchangeableCArray[exchangeableCIdx]] = true;
				++exchangeableCIdx;
			}
			else
			{
				servers.push_front(lastChangedNode); // restore the exchangeable node removed in previous exchange operation
				deployed[lastChangedNode] = true;
				progress = Progress::EX2; // return back to Progress::EX2 and try the next exchangeable node
			}
		}
		else
		{
			lastResultFee = justResultFee;

			info_log("node [%d] <-> [%d] in Progress::EX2  \t---  %ld\n", lastChangedNode, exchangeableCArray[exchangeableCIdx-1], justResultFee);

			inExchangeProgress = false;
			progress = Progress::DL; // go to Progress::DL and try to decrease server's level in new scheme
		}
		break;
	}
	case Progress::RM:
	{
		if (!inRemovingProgress)
		{
			inRemovingProgress = true;
			info_log("====================    Progress::RM starts    ====================\n");

			std::vector<std::pair<double /* unit flux fee */, int /* node ID */> > sortUnitFluxFee;
			sortUnitFluxFee.reserve(servers.size());
			for (itS = servers.begin(), i = 0; itS != servers.end(); ++itS, ++i)
			{
				int curLevel = fitLevel(serversOutputFlux[i]);
				double totalNodeFee = priceTable[curLevel] + deployTable[*itS];
				sortUnitFluxFee.emplace_back(totalNodeFee/serversOutputFlux[i], *itS);
				// printf("%d[%.2f] ", *itS, totalNodeFee/serversOutputFlux[i]);
			}
			// printf("\n");
			std::sort(sortUnitFluxFee.begin(), sortUnitFluxFee.end(), std::greater<std::pair<double, int> >());
			removableNum = std::min(kRemovableNum, static_cast<uint32_t>(sortUnitFluxFee.size()));
			for (i = 0; i < removableNum; ++i)
			{
				removableArray[i] = sortUnitFluxFee[i].second;
				// printf("%f %d %d\n", sortUnitFluxFee[i].first, removableArray[i], scanner.nodeLevel[removableArray[i]]);
			}
			removableIdx = 0;
			lastChangedNode = removableArray[removableIdx++];
			REMOVE_CHANGED_NODE
			deployed[lastChangedNode] = false;
		}
		else
		{
			if (lastResultFee <= justResultFee || justRentalFee == -1) // try to restore the removable node
			{
				servers.push_front(lastChangedNode);
				deployed[lastChangedNode] = true;
				if (removableIdx < removableNum)
				{
					lastChangedNode = removableArray[removableIdx++];
					REMOVE_CHANGED_NODE
					deployed[lastChangedNode] = false;
				}
				else
				{
					printf("====================    Exit greedy category    ====================\n\n");
					return false;
				}
			}
			else
			{
				lastResultFee = justResultFee;

				info_log("node [%d] is successfully removed in Progress::RM  ---  %ld\n", lastChangedNode, justResultFee);

				inRemovingProgress = false;
				progress = Progress::DL; // go to Progress::DL and try to decrease server's level in new scheme
			}
		}
		break;
	}
	default:
		error_log("Unknown progress!\n");
	}

	return true;
}
