/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: Utils.h                                                                    *
 * @author: Xu Le                                                                    *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#ifndef __UTILS_H__
#define __UTILS_H__

#include <cstdint>
#include <cstring>
#include <vector>
#include <list>
#include <stack>
#include <queue>
#include <set>
#include <unordered_set>
#include <unordered_map>
#include <algorithm>
#include <utility>

#include "Timer.h"

#define square(x)    ((x)*(x))

/** structure used to store input lines' parsed result. */
struct LinkTuple
{
	LinkTuple(int _src, int _dst, int _bandwidth, int _rental) : src(_src), dst(_dst), bandwidth(_bandwidth), rental(_rental) {}

	LinkTuple(const LinkTuple& rhs) = default;
	LinkTuple& operator=(const LinkTuple& rhs) = default;

	int src;       ///< first element in a single input line.
	int dst;       ///< second element in a single input line.
	int bandwidth; ///< link bandwidth.
	int rental;    ///< link rental fee.
};

/** case complexity enum value space. */
enum class Complexity {
	Oversimplified,
	Simple,
	General,
	Hard,
	VeryHard,
	UltraHard,
	Hell
};

/** parse input lines from case file. */
void parseInput(char *filename, std::list<struct LinkTuple>& linkTuples);

/** output flux path list to result file. */
void outputToFile(char *filename, std::list<std::vector<int> >& fluxPathList);

/** output NA to result file. */
void outputToFile(char *filename);

/** binary search in a certain std::vector, cost O(log n) */
bool binarySearch(const std::vector<int>& vec, const int key);

/** determine the highest server level to be used. */
void defineAvailableLevel();

/**
 * locate the server level of which a network node should deployed according to node's connected link capacity,
 * interval of nodeCap locate in [ oupcapTable[x], oupcapTable[x+1] ), note the function return 0
 * when nodeCap < outcapTable[0], cost O(log n).
 */
int locateLevel(const int nodeCap);

/**
 * fit the server level of which a network node should deployed according to node's actual output flux obtained by MCF,
 * interval of outputFlux locate in ( oupcapTable[x], oupcapTable[x+1] ], note the function return hardwareLevelNum
 * when nodeCap > outcapTable[hardwareLevelNum-1], cost O(log n).
 */
int fitLevel(const int outputFlux);

/** obtain the total purchase fee under the condition that all chosen network nodes deploy fitted server level. */
long obtainTotalPurchaseFee(std::vector<int>& serversOutputFlux, size_t serverNum);

/** define the default complexity of the case according to its size. */
Complexity defineDefaultComplexity();

/** convert enum value to its name string, only useful for debuging purpose. */
const char* complexityString(Complexity complexity);

////////////////    convenient functions of std::vector and std::list    ////////////////
void printVector(std::vector<int>& _vector, const char *_note);

void printVector(std::vector<double>& _vector, const char *_note);

void printList(std::list<int>& _list, const char *_note);

void printList(std::list<double>& _list, const char *_note);

#endif /* __UTILS_H__ */
